import { ChainName, ChainToPlatform, Network, PlatformName } from "@wormhole-foundation/sdk-base";
import { ChainAddress, ChainContext, Contracts, NativeAddress, PayloadDiscriminator, PayloadLiteral, Platform, TokenId, TxHash, UniversalAddress, WormholeMessageId, deserialize } from "@wormhole-foundation/sdk-definitions";
import { WormholeConfig } from "./types";
import { CircleTransfer } from "./protocols/cctpTransfer";
import { TokenTransfer } from "./protocols/tokenTransfer";
import { TransactionStatus } from "./whscan-api";
export declare class Wormhole {
    protected _platforms: Map<PlatformName, Platform<PlatformName>>;
    protected _chains: Map<ChainName, ChainContext<ChainToPlatform<ChainName>>>;
    protected readonly _network: Network;
    readonly config: WormholeConfig;
    constructor(network: Network, platforms: Platform<PlatformName>[], config?: WormholeConfig);
    get network(): Network;
    /**
     * Creates a CircleTransfer object to move Native USDC from one chain to another
     * @param amount the amount to transfer
     * @param from the address to transfer from
     * @param to the address to transfer to
     * @param automatic whether to use automatic delivery
     * @param payload the payload to send with the transfer
     * @param nativeGas the amount of native gas to send with the transfer
     * @returns the CircleTransfer object
     * @throws Errors if the chain or protocol is not supported
     */
    circleTransfer(amount: bigint, from: ChainAddress, to: ChainAddress, automatic: boolean, payload?: Uint8Array, nativeGas?: bigint): Promise<CircleTransfer>;
    /**
     * Creates a TokenTransfer object to move a token from one chain to another
     * @param token the token to transfer
     * @param amount the amount to transfer
     * @param from the address to transfer from
     * @param to the address to transfer to
     * @param automatic whether to use automatic delivery
     * @param payload the payload to send with the transfer
     * @param nativeGas the amount of native gas to send with the transfer
     * @returns the TokenTransfer object
     * @throws Errors if the chain or protocol is not supported
     */
    tokenTransfer(token: TokenId | "native", amount: bigint, from: ChainAddress, to: ChainAddress, automatic: boolean, payload?: Uint8Array, nativeGas?: bigint): Promise<TokenTransfer>;
    /**
     * Gets the contract addresses for a given chain
     * @param chain the chain name or chain id
     * @returns the contract addresses
     */
    getContracts(chain: ChainName): Contracts | undefined;
    /**
     * Returns the platform object, i.e. the class with platform-specific logic and methods
     * @param chain the chain name or platform name
     * @returns the platform context class
     * @throws Errors if platform is not found
     */
    getPlatform(chain: ChainName | PlatformName): Platform<PlatformName>;
    /**
     * Returns the chain "context", i.e. the class with chain-specific logic and methods
     * @param chain the chain name
     * @returns the chain context class
     * @throws Errors if context is not found
     */
    getChain<CN extends ChainName>(chain: CN): ChainContext<ChainToPlatform<CN>>;
    /**
     * Gets the TokenId for a token representation on any chain
     *  These are the Wormhole wrapped token addresses, not necessarily
     *  the cannonical version of that token
     *
     * @param tokenId The Token ID (chain/address)
     * @param chain The chain name or id
     * @returns The TokenId on the given chain, null if it does not exist
     * @throws Errors if the chain is not supported or the token does not exist
     */
    getWrappedAsset(chain: ChainName, token: TokenId): Promise<TokenId>;
    /**
     * Gets the number of decimals for a token on a given chain
     *
     * @param tokenId The Token ID (home chain/address)
     * @param chain The chain name or id of the token/representation
     * @returns The number of decimals
     */
    getDecimals(chain: ChainName, token: NativeAddress<PlatformName> | UniversalAddress | "native"): Promise<bigint>;
    /**
     * Converts a human friendly decimal number to base units for the token passed
     *
     * @param chain The chain name
     * @param tokenId The token ID (its home chain and address on the home chain) or 'native'
     * @param amount The decimal number as a string to convert into base units
     * @returns The amount converted to base units as a BigNumber
     */
    normalizeAmount(chain: ChainName, token: UniversalAddress | NativeAddress<PlatformName> | "native", amount: number | string): Promise<bigint>;
    /**
     * Fetches the balance of a given token for a wallet
     *
     * @param walletAddress The wallet address
     * @param tokenId The token ID (its home chain and address on the home chain)
     * @param chain The chain name or id
     * @returns The token balance of the wormhole asset as a BigNumber
     */
    getBalance(chain: ChainName, token: NativeAddress<PlatformName> | UniversalAddress | "native", walletAddress: string): Promise<bigint | null>;
    /**
     * Gets the associated token account for chains that require it (only Solana currently).
     * @param sendingChain the chain name of the source token
     * @param sendingToken the TokenId or address of the source token
     * @param recipient the address of the recipient
     * @returns
     */
    getTokenAccount(sendingChain: ChainName, sendingToken: UniversalAddress | NativeAddress<PlatformName> | TokenId | "native", recipient: ChainAddress): Promise<ChainAddress>;
    /**
     * Gets the Raw VAA Bytes from the API or Guardian RPC, finality must be met before the VAA will be available.
     * @param chain The chain name
     * @param emitter The emitter address
     * @param sequence The sequence number
     * @param retries The number of times to retry
     * @param opts The options for the request. Timeouts must be set in ms.
     * @returns The VAA bytes if available
     * @throws Errors if the VAA is not available after the retries
     */
    getVaaBytes(chain: ChainName, emitter: UniversalAddress | NativeAddress<PlatformName>, sequence: bigint, timeout?: number): Promise<Uint8Array | undefined>;
    /**
     * Gets a VAA from the API or Guardian RPC, finality must be met before the VAA will be available.
     * @param chain The chain name
     * @param emitter The emitter address
     * @param sequence The sequence number
     * @param timeout The total amount of time to wait for the VAA to be available
     * @returns The VAA if available
     * @throws Errors if the VAA is not available after the retries
     */
    getVaa<T extends PayloadLiteral | PayloadDiscriminator>(chain: ChainName, emitter: UniversalAddress | NativeAddress<PlatformName>, sequence: bigint, decodeAs: T, timeout?: number): Promise<ReturnType<typeof deserialize<T>> | undefined>;
    getCircleAttestation(msgHash: string, timeout?: number): Promise<string | null>;
    /**
     * Get the status of a transaction, identified by the chain, emitter address, and sequence number
     *
     * @param chain the chain name
     * @param emitter the emitter address
     * @param sequence the sequence number
     * @returns the TransactionStatus
     */
    getTransactionStatus(chain: ChainName, emitter: UniversalAddress | NativeAddress<PlatformName>, sequence: bigint, timeout?: number): Promise<TransactionStatus>;
    /**
     * Parse an address to a universal address
     *
     * @param address The native address
     * @returns The address in the universal format
     */
    parseAddress(chain: ChainName, address: string): NativeAddress<PlatformName>;
    /**
     * Parses all relevant information from a transaction given the sending tx hash and sending chain
     *
     * @param tx The sending transaction hash
     * @param chain The sending chain name or id
     * @returns The parsed WormholeMessageId
     */
    parseMessageFromTx(chain: ChainName, txid: TxHash, timeout?: number): Promise<WormholeMessageId[]>;
}
//# sourceMappingURL=wormhole.d.ts.map